import signal
import contextlib
from multiprocessing import parent_process


@contextlib.contextmanager
def graceful_interrupt_handler(
    signums=[signal.SIGINT, signal.SIGQUIT],
    callback=lambda _: None,
):
    for signum in signums:
        signal.signal(signum, lambda signum, _: parent_process() or callback(signum))
    yield
    for signum in signums:
        signal.signal(signum, signal.SIG_DFL)



from pathlib import Path
from natsort import natsorted


class EMAMeter:
    def __init__(self, decay=0.98):
        self.decay = decay
        self.value = None

    def __call__(self, value):
        if self.value is None:
            self.value = value
        else:
            self.value = self.decay * self.value + (1 - self.decay) * value
        return self.value


def print_directory_tree(root: Path, prefix: str = ""):
    if not root.exists():
        return
    print(f"{prefix}{root.name if prefix else root}")
    if root.is_dir():
        base = prefix.replace("─", " ").replace("├", "│").replace("└", " ")
        paths = natsorted(root.iterdir())
        for i, path in enumerate(paths):
            if i < len(paths) - 1:
                print_directory_tree(path, base + "├── ")
            else:
                print_directory_tree(paths[-1], base + "└── ")



import torch
from queue import Queue
from pathlib import Path
from functools import lru_cache


def load_state_dict_non_strict(model, state_dict):
    model_state_dict = model.state_dict()
    provided = set(state_dict)
    required = set(model_state_dict)
    agreed = provided & required
    for k in list(agreed):
        if model_state_dict[k].shape != state_dict[k].shape:
            agreed.remove(k)
            provided.remove(k)
    state_dict = {k: state_dict[k] for k in agreed}
    if (diff := provided - required) :
        print("Provided but not required keys: ")
        print(diff)
    if (diff := required - provided) :
        print("Required but not provided keys: ")
        print(required - provided)
    model.load_state_dict(state_dict, strict=False)
    return model


@lru_cache
def cached_load_ckpt(path):
    return torch.load(path, "cpu")


class Buffer(dict):
    def __init__(self, get_path):
        self.get_path = get_path


class Saver:
    def __init__(self, root: Path, strict: bool = True, buffer_size=1):
        self.root = root
        self.strict = strict
        self.buffered = Queue(buffer_size)

    @property
    def ckpts(self):
        return list(self.root.glob("*.ckpt"))

    @property
    def empty(self):
        return len(self.ckpts) > 0

    @property
    def latest_ckpt(self):
        return max(self.ckpts, key=lambda ckpt: self.parse(ckpt)[1], default=None)

    def get_path(self, model):
        return self.root / f"epoch={model.epoch}-iteration={model.iteration}.ckpt"

    def parse(self, path):
        epoch, iteration = map(lambda kv: int(kv.split("=")[1]), path.stem.split("-"))
        return epoch, iteration

    def load(self, ckpt, model=None, optimizers=[], scaler=None):
        if ckpt is None:
            return

        state_dict = cached_load_ckpt(ckpt)
        epoch, iteration = state_dict["epoch"], state_dict["iteration"]

        if model is not None:
            if self.strict:
                load_state_dict_non_strict(model, state_dict["model"])
            else:
                model.load_state_dict(state_dict["model"])
            model.epoch, model.iteration = epoch, iteration
            print(f"Model epoch={epoch} iteration={iteration} loaded.")

        for i, (optimizer, optimizer_state_dict) in enumerate(
            zip(optimizers, state_dict.get("optimizers", []))
        ):
            try:
                optimizer.load_state_dict(optimizer_state_dict)
                # sanity check
                optimizer.zero_grad()
                optimizer.step()
                print(f"Optimizer at epoch {epoch} loaded.")
            except Exception as e:
                print(e)
                print(f"Warning: fail to load state dict for optimizer {i}.")

        if scaler is not None:
            try:
                scaler.load_state_dict(state_dict["scaler"])
                print(f"Scaler at epoch {epoch} loaded.")
            except Exception as e:
                print(e)
                print("Warning: loading scaler state dict failed.")

    def buffer(self, model, optimizers=[], scaler=None):
        path = self.get_path(model)
        data = dict(
            iteration=model.iteration,
            epoch=model.epoch,
            model=model.state_dict(),
            optimizers=[optimizer.state_dict() for optimizer in optimizers],
            scaler=scaler.state_dict() if scaler else None,
        )
        if self.buffered.full():
            self.buffered.get()
        self.buffered.put([path, data])

    def dump(self):
        while not self.buffered.empty():
            path, data = self.buffered.get()
            path.parent.mkdir(parents=True, exist_ok=True)
            torch.save(data, path)
            print(f"{path} saved.")

    def save(self, model, optimizers=[], scaler=None):
        self.buffer(model, optimizers, scaler)
        self.dump()



import math
import time
import tqdm
import numbers
from collections import defaultdict

from .utils import EMAMeter


class ProgressBar(tqdm.tqdm):
    def __init__(self, iterable, dynamic_ncols=True, mininterval=0.5, **kwargs):
        super().__init__(
            iterable,
            dynamic_ncols=dynamic_ncols,
            mininterval=mininterval,
            **kwargs,
        )
        self.mininterval = mininterval
        self.lines = defaultdict(self.create_line)
        self.emas = defaultdict(EMAMeter)

    def timeup(self):
        return time.time() - self.last_print_t > self.mininterval

    def create_line(self):
        return tqdm.tqdm(bar_format="╰ {unit}")

    def update_line(self, k, v, fmt=None):
        if isinstance(v, numbers.Number) and math.isfinite(v):
            v = self.emas[k](v)
            fmt = fmt or "{k}: {v:.4g}"
        else:
            fmt = fmt or "{k}: {v}"
        first_update = k not in self.lines
        line = self.lines[k]
        line.unit = fmt.format(k=k, v=v)
        if first_update:
            line.refresh()
        elif self.timeup():
            line.refresh()

    def close(self):
        super().close()
        for line in self.lines.values():
            line.close()



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="torchzq",
    python_requires=">=3.6.0",
    version="1.0.8",
    description="TorchZQ: A simple PyTorch experiment runner.",
    author="enhuiz",
    author_email="niuzhe.nz@outlook.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    scripts=["tzq"],
    install_requires=[
        "torch",
        "zouqi==1.0.8",
        "pyyaml",
        "tqdm",
        "pandas",
        "torchvision",
        "natsort",
        "deprecated",
    ],
    url="https://github.com/enhuiz/torchzq",
)



import time
import yaml
import sys
import os
import shutil
import numbers
import numpy as np
import random
import torch
import torch.nn as nn
import signal
from pathlib import Path
from collections import defaultdict
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter
from functools import partial
from collections import defaultdict

import zouqi

from .typing import Flag, Scheduled, _Scheduled
from .saver import Saver
from .scheduler import Scheduler
from .pbar import ProgressBar
from .utils import print_directory_tree
from .interrupt import graceful_interrupt_handler


def immutable(name):
    def getter(self):
        return getattr(self, "__" + name, None)

    def setter(self, x):
        assert getter(self) == None, "Immutable variable cannot be set twice!"
        setattr(self, "__" + name, x)

    return property(getter, setter)


def delegate(delegate_cls, name):
    attrs = set(delegate_cls.__dict__.keys())

    def wrapped(cls):
        nonexist_attrs = attrs - set(cls.__dict__.keys())
        for attr in nonexist_attrs:
            getsrc = lambda self: getattr(self, name)
            getter = lambda self, attr=attr: getattr(getsrc(self), attr)
            setter = lambda self, x, attr=attr: setattr(getsrc(self), attr, x)
            setattr(cls, attr, property(getter, setter))
        return cls

    return wrapped


class Mode(str):
    logger = immutable("logger")
    data_loader = immutable("data_loader")


@delegate(Mode, "mode")
class Runner:
    model = immutable("model")
    saver = immutable("saver")
    optimizers = immutable("optimizers")
    scheduler = immutable("scheduler")
    scaler = immutable("scaler")
    args = immutable("args")

    def __init__(
        self,
        name: str = "Unnamed",
        batch_size: int = 32,
        nj: int = min(os.cpu_count(), 12),
        device: str = "cuda",
        strict_loading: bool = True,
        runs_root: Path = Path("runs"),
        quiet: Flag = False,
        use_fp16: bool = False,
        ckpt: Path = None,
        lr: Scheduled = "1e-3",
        from_scratch: Flag = False,
        seed: int = 0,
    ):
        self.modes = []
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)

    @property
    def mode(self):
        if len(self.modes) == 0:
            raise RuntimeError('No mode has been set, call "self.switch_mode()" first.')
        return self.modes[0]

    @property
    def name(self):
        return self.args.name

    @property
    def dataset(self):
        return self.data_loader.dataset

    @property
    def training(self):
        return self.mode == "train"

    @property
    def ckpt(self):
        args = self.args
        if args.ckpt is not None:
            return args.ckpt
        if not args.from_scratch:
            return self.saver.latest_ckpt
        return None

    @property
    def run_dir(self):
        return self.args.runs_root / self.name

    @property
    def logs_dir(self):
        return self.run_dir / "logs" / self.mode

    @property
    def ckpt_dir(self):
        return self.run_dir / "ckpts"

    @property
    def events(self):
        return self.mode.events

    @property
    def autocast_if_use_fp16(self):
        return partial(torch.cuda.amp.autocast, enabled=self.args.use_fp16)

    @property
    def batch_size(self):
        return self.args.batch_size

    @property
    def Dataset(self):
        raise NotImplementedError

    @property
    def DataLoader(self):
        return DataLoader

    @staticmethod
    def set_lr(optimizer, lr):
        for g in optimizer.param_groups:
            g["lr"] = lr

    @staticmethod
    def get_lr(optimizer):
        return optimizer.param_groups[0]["lr"]

    def create_scheduler(self):
        scheduler = Scheduler()
        for k, v in list(vars(self.args).items()):
            if isinstance(v, _Scheduled):
                setattr(self.args, k, scheduler.schedule(v))
        return scheduler

    def create_dataset(self):
        raise NotImplementedError

    def create_data_loader(self, **kwargs):
        args = self.args
        dataset = self.create_dataset()
        data_loader = self.DataLoader(
            dataset=dataset,
            batch_size=self.batch_size,
            num_workers=args.nj,
            **kwargs,
        )
        print("Dataset size:", len(dataset))
        return data_loader

    def create_model(self):
        raise NotImplementedError

    def create_optimizers(self):
        args = self.args
        optimizer = torch.optim.Adam(params=self.model.parameters(), lr=1)
        args.lr.add_listeners(lambda lr: self.set_lr(optimizer, lr))
        return [optimizer]

    def prepare_saver(self):
        args = self.args
        if self.saver is None:
            self.saver = Saver(self.ckpt_dir, args.strict_loading)

    def prepare_logger(self):
        if self.logger is None:
            self.logger = SummaryWriter(self.logs_dir)

    def prepare_data_loader(self):
        # data_loader should be before the model
        if self.data_loader is None:
            self.data_loader = self.create_data_loader(
                shuffle=self.training, drop_last=self.training
            )

    def prepare_scheduler(self):
        # scheduler should be before the model
        if self.scheduler is None:
            self.scheduler = self.create_scheduler()

    def prepare_model(self):
        args = self.args
        if self.model is None:
            self.model = self.create_model()
            self.model.to(args.device)
            self.model.epoch = 0
            self.model.iteration = 0
            self.saver.load(self.ckpt, model=self.model)
            self.scheduler.step(epoch=self.model.epoch, iteration=self.model.iteration)

    def prepare_optimizer(self):
        args = self.args
        if self.optimizers is None:
            self.optimizers = self.create_optimizers()
            if len(self.optimizers) == 0:
                raise ValueError("There should be at least 1 optimizer but get 0.")
            for optimizer in self.optimizers:
                for state in optimizer.state.values():
                    for k, v in state.items():
                        if torch.is_tensor(v):
                            state[k] = v.to(args.device)
            self.saver.load(self.ckpt, optimizers=self.optimizers)

    def prepare_scaler(self):
        args = self.args
        if self.scaler is None and args.use_fp16:
            self.scaler = torch.cuda.amp.GradScaler()
            self.saver.load(self.ckpt, scaler=self.scaler)

    @staticmethod
    def run_pipeline(*stages):
        for stage in stages:
            for func in stage:
                func()

    def prepare_all(self):
        self.run_pipeline(
            [self.prepare_saver, self.prepare_scheduler],
            [self.prepare_data_loader],
            [self.prepare_model],
            [self.prepare_optimizer, self.prepare_scaler],
            [self.prepare_logger],
        )

    def switch_mode(self, name):
        """
        Switch running mode.
        """
        # the first element in self.modes will be treated as the current mode
        if name in self.modes:
            i = self.modes.index(name)
            self.modes[0], self.modes[i] = self.modes[i], self.modes[0]
        else:
            self.modes.insert(0, Mode(name))
            self.prepare_all()
        if self.training:
            self.model.train()
        elif self.model is not None:
            self.model.eval()

    def training_step(self, batch, optimizer_idx: int) -> tuple[torch.Tensor, dict]:
        raise NotImplementedError

    def validation_step(self, batch, batch_idx: int) -> dict:
        stat_dict = {}
        for i in range(len(self.optimizers)):
            stat_dict.update(self.training_step(batch, i)[1])
        return stat_dict

    def testing_step(self, batch, batch_idx: int) -> dict:
        raise NotImplementedError

    def training_step_with_optimization(self, batch) -> dict:
        args = self.args
        model = self.model
        stat_dict = {}
        start_time = time.time()

        model.iteration += 1
        self.scheduler.step(model.epoch, model.iteration)

        for optimizer_idx, optimizer in enumerate(self.optimizers):
            with self.autocast_if_use_fp16():
                outputs = self.training_step(batch, optimizer_idx)

            if isinstance(outputs, torch.Tensor):
                loss = outputs
            else:
                loss = outputs[0]
                stat_dict.update(outputs[1])

            if args.use_fp16:
                self.scaler.scale(loss / args.update_every).backward()
            else:
                (loss / args.update_every).backward()

            if model.iteration % args.update_every == 0:
                if args.use_fp16:
                    self.scaler.unscale_(optimizer)

                grad_norm = nn.utils.clip_grad_norm_(
                    self.model.parameters(),
                    args.grad_clip_thres or 1e9,
                )

                stat_dict[f"grad_norm_{optimizer_idx}"] = grad_norm.item()

                if args.use_fp16:
                    self.scaler.step(optimizer)
                    self.scaler.update()
                else:
                    optimizer.step()

                optimizer.zero_grad()

        stat_dict["elapsed_time"] = time.time() - start_time

        return stat_dict

    def training_loop(self):
        if not self.training:
            print("Warning: You are calling training loop in non-training mode!")

        args = self.args
        model = self.model
        logger = self.logger

        while model.epoch < args.max_epochs:
            model.epoch += 1
            desc = f"Train: {model.epoch}/{args.max_epochs}"
            pbar = ProgressBar(self.data_loader, args.quiet, desc=desc)

            def interrupt_callback(signum):
                pbar.close()
                print("Trying to gracefully shutdown ...")
                self.saver.dump()
                if signum == signal.SIGQUIT:
                    self.saver.save(model, self.optimizers, self.scaler)
                sys.exit(0)

            with graceful_interrupt_handler(callback=interrupt_callback):
                for batch in pbar:
                    try:
                        stat_dict = self.training_step_with_optimization(batch)
                        pbar.update_line(0, f"iteration: {model.iteration}", "{v}")
                        for key, val in stat_dict.items():
                            pbar.update_line(key, val)
                        if model.iteration % args.log_every == 0:
                            for key, val in stat_dict.items():
                                logger.add_scalar(key, val, model.iteration)
                            logger.flush()
                    except RuntimeError as e:
                        if "out of memory" in str(e):
                            pbar.update_line(0, "OOM! Skip batch.", "{v}")
                            for p in model.parameters():
                                if p.grad is not None:
                                    del p.grad  # free some memory
                            torch.cuda.empty_cache()
                            continue
                        else:
                            raise e
                pbar.close()
                self.saver.buffer(model, self.optimizers, self.scaler)
                if model.epoch % args.save_every == 0:
                    self.saver.dump()
                if model.epoch % args.validate_every == 0:
                    self.validate()
                    self.switch_mode("train")

    def val_test_loop(self, desc, step, sanity_check=False):
        args = self.args
        model = self.model
        data_loader = self.data_loader
        if sanity_check:
            data_loader = [batch for batch, _ in zip(data_loader, range(3))]
        pbar = ProgressBar(data_loader, args.quiet, desc=desc)
        stats_dict = defaultdict(list)
        for index, batch in enumerate(pbar):
            stat_dict = step(batch, index)
            for key, val in (stat_dict or {}).items():
                pbar.update_line(key, val)
                if isinstance(val, numbers.Number):
                    stats_dict[key].append(val)
        for key, val in stats_dict.items():
            mean = sum(val) / len(val)
            self.logger.add_scalar(key, mean, model.epoch)
            print(f"Average {key}: {mean:.4g}.")
        self.logger.flush()

    @zouqi.command
    def train(
        self,
        weight_decay: float = 0,
        max_epochs: int = 100,
        save_every: int = 1,
        validate_every: int = None,
        update_every: int = 1,
        log_every: int = 10,
        grad_clip_thres: float = 1.0,
    ):
        args = self.args
        if validate_every is None:
            args.validate_every = save_every

        self.run_dir.mkdir(parents=True, exist_ok=True)
        time_str = time.strftime("%Y%m%dT%H%M%S")
        with open(self.run_dir / f"config-{time_str}.yml", "w") as f:
            yaml.dump(vars(args), f)

        self.switch_mode("validate")
        self.val_test_loop(
            "Validation loop sanity checking ...",
            self.validation_step,
            sanity_check=True,
        )
        self.switch_mode("train")
        self.training_loop()

    @zouqi.command
    def validate(self):
        self.switch_mode("validate")
        self.val_test_loop(
            f"Validating epoch {self.model.epoch} ...",
            self.validation_step,
        )

    @zouqi.command
    def test(self):
        self.switch_mode("test")
        self.val_test_loop(
            f"Testing epoch {self.model.epoch} ...",
            self.testing_step,
        )

    @staticmethod
    def try_rmtree(path):
        if path.exists():
            shutil.rmtree(path)
            print(str(path), "removed.")

    @zouqi.command
    def clear(self):
        args = self.args
        self.ls()
        if (args.runs_root / self.name).exists():
            if input("Are you sure to clear? (y)\n").lower() == "y":
                self.try_rmtree(args.runs_root / self.name)
            else:
                print(f"Not cleared.")
        else:
            print("Nothing to clear.")

    @zouqi.command
    def ls(self):
        args = self.args
        print_directory_tree(args.runs_root / self.name)
